#!/bin/bash
########################################################################
# build_tbb.sh
#
# Build tbb and package for UPS.
#
# Adapted from build.sh.template provided by ssibuildshims v2_09_00.
########################################################################

########################################################################
# Non-customize-able boilerplate: skip to "Customize-able functions."
########################################################################

####################################
# Useful variables.
prog=${BASH_SOURCE[0]##*/}

##################
# Local (dumb) ssi_die() function for use until we source ssi_functions.
ssi_die() {
  printf "FATAL ERROR: ${*}\n" 1>&2; exit 1
}
##################

##################
# Basic usage function for use by ensure_buildshims until we have enough
# information to provide something more specific. Do NOT customize!
build_script_usage() {
    cat <<EOF
USAGE: ${prog} <ssi-msg-config-options> [--] <product-dir> <qualifier-args> <maketar>
       ${prog} -h|--help [<product-dir>]

  Build ${package} and package for UPS.

OPTIONS

  -h|--help

    Help. Supply <product-dir> for specific, detailed information on
    invoking ${prog}.

  <ssi-msg-config-options>

    Supply <product-dir> for more information.

ARGUMENTS

  <product-dir>

    The top-level UPS directory in which the source code distribution
    was unpacked. When specified with -h|--help, this enables us to
    provide more specific help about invoking ${prog}.

  <qualifier-args>
  <maketar>

    Supply <product-dir> for more information.

EOF
}
##################

##################
# Make sure we have UPS and ssibuildshims set up.
ensure_ssibuildshims() {
  if ! (( $# )) || [ "${@: -1}" = "-h" ] || [ "${@: -1}" = "--help" ]; then
    build_script_usage
    exit 1
  fi
  # Find <product-dir> from our arguments:
  local arg
  for arg in "${@}"; do
    [[ "$arg" ==  -* ]] && continue;
    product_dir="${arg}"; break
  done
  if [ -z "${product_dir}" ]; then
    # Bad usage.
    ssi_die "Bad <product-dir> ${product_dir}\n$(build_script_usage)"
    exit 2
  elif ! { [ -d "${product_dir}/.upsfiles" ] && [ -w "${product_dir}" ]; }; then
    # Something wrong with <product-dir>.
    ssi_die "<product-dir> not a valid, writable unified UPS directory: " \
      "${product_dir}\n$(build_script_usage)"
  fi
  # Promising...
  local abs_product_dir="$(cd "${product_dir}" 2>/dev/null 2>&1 && pwd -P)"
  (( $? )) && ssi_die "unable to cd to <product-dir>: ${product_dir}\n" \
    "$(build_script_usage)"
  # Find our own bootstraps and yank 'em.
  local ssibuildshims_topdir="${product_dir}/ssibuildshims/${ssibuildshims_version}"
  [ -f "${ssibuildshims_topdir}/bin/ssi_functions" ] && \
    source "${ssibuildshims_topdir}/bin/ssi_functions" >/dev/null 2>&1 && \
    ensure_ups "${product_dir}" && \
    setup ssibuildshims ${ssibuildshims_version} -z "${product_dir}" || \
    ssi_die "unable to find ssibuildshims ${ssibuildshims_version} in
${product_dir} and set it up with UPS."
  # Absolute is safer.
  absolute_path -- "${product_dir}" product_dir
}
##################

########################################################################
# Customize-able functions.
########################################################################

##################
# Ascertain whether we built and installed correctly. Add more /
# different calls to verify_in_filesystem as appropriate. Arguments
# (e.g. -q) are passed through to verify_in_filesystem.
is_built() { # *C*
  ${trace} verify_in_filesystem -d "${@}" "${pkgdir}/lib"{64,}
}
##################

##################
# If a distribution archive is wanted and the installation is successful
# according to is_built(), then make it. Any arguments (e.g. -q) are
# passed through both to is_built() and make_basic_distribution().
#
# If make_basic_distribution() finds that the product is declared
# correctly and attempts to make the distribution, then it will cause us
# to exit with the status code of that operation. If not, it returns an
# error code.
#
# See make_basic_distribution -h for useful options to add to
# mbd_opts(). *C*
maybe_make_distribution() {
  if [ "${maketar}" = "tar" ]; then
    # Customize mbd_opts. *C*
    local mbd_opts=()
    ${trace} is_built "$@" && \
      ${trace} make_basic_distribution "${@}" "${mbd_opts[@]}"
  fi
}
##################

########################################################################
# Main.
########################################################################

####################################
# Useful variables. *C*
trace=maybe_xtrace # Could set to xtrace instead for unconditional.
(( ssi_trace = 1 )) # Trace certain commands by default.
####################################

####################################
# Package name and version information. *C*
package=tbb
origpkgver=v2021_1_1
pkgver=${origpkgver} # Append letters for local patches.
ssibuildshims_version=v2_09_00
pkgdotver=${origpkgver//_/.}; pkgdotver=${pkgdotver#v}
####################################

##################
# Other useful variables. *C*
pkgtarname=oneTBB-${pkgdotver}
pkgtarfile=${pkgtarname}.tar.gz
##################

####################################
# Handle options and arguments.

# Set up ssibuildshims and UPS / product_dir as early as possible.
ensure_ssibuildshims "${@}"

##################
# Now parse options and arguments, setting basequal, extraqual, maketar,
# etc.
#
# Use verify_quals() to parse command line options, verify that the user
# has provided qualifiers supported by this build script, and set
# expected variables such as basequal, extraqual, cqual, etc.
#
# Customize the extra usage text as appropriate. Note that DEBUG_PATCHES
# and VERBOSE_PATCHES are honored by prepare_source() and
# patch_source(). DEBUG_(CONFIG|BUILD|TEST|INSTALL) should be accounted
# for (or not) by the relevant command in this script.
#
# See verify_quals -h for options and examples.
verify_quals \
  --cqual={e19,e20,e21,c7} \
  --parse --usage-file=- -- "$@" <<EOF || exit # *C*
ENVIRONMENT

  DEBUG_(CONFIG|BUILD)

    Verbose output from the corresponding step.

  DEBUG_PATCHES
  VERBOSE_PATCHES

    Command tracing and increased verbosity for source unpacking and patching
    operations. DEBUG_PATCHES will cause exit after patches have been applied.
EOF
##################

##################
# Define globals: pkgdir, sourcedir, blddir, fullqual, etc. See
# define_basics -h for steering options.
${trace} define_basics NO_DEBUG_SOURCE NO_TYPE TRACK_PROGRESS || exit # *C*
##################

####################################
# Declare ourselves privately, and then set ourselves up to allow setup
# of dependencies. *C*
#
# Comment or delete if not required:
${trace} fake_declare_product "${ssi_args[@]}" -x -- || exit
####################################

####################################
# Build-only dependencies. *C*
${trace} setup_cmake --latest={--ge,3.15} || exit
${trace} setup_ninja --optional && (( want_ninja = 1 ))
####################################

# Print active products.
echo
ups active
echo

# Possible short-circuit if we're already built and declared.
${trace} maybe_make_distribution --silent

ssi_info "building ${package} for ${OS}-${plat}-${qualdir} (flavor ${pkgflvr})"

# Ensure directories: pkgdir, sourcedir, blddir.
${trace} init_pkg_dirs || exit

####################################
# Prepare source for building. *C*

${trace} cd "${sourcedir}" \
  || ssi_die "unable to cd to source directory ${sourcedir}"

# Source tarballs and patches. See prepare_source -h / patch_source -h
# for options.
${trace} prepare_source "${ssi_args[@]}" -P "${patchdir}" \
  "${tardir}/${pkgtarfile}" \
  || exit

# Bail early if DEBUG_PATCHES is set.
[ -n "${DEBUG_PATCHES}" ] && exit 1
####################################

####################################
# Prepare to build. *C*
#
# Set configure / compile flags, environment variables, etc.
set -x
cmake_options=()
cxxflg=(-fno-math-errno)

cc=gcc; cxx=g++; cid=GNU
case ${cqual} in
  e19|e20) cxxstd=17;;
  e21) cxxstd=20;;
  c7) cc=clang; cxx=clang++; cid=Clang; cxxstd=17;;
  *) ssi_die -I "accepted unknown cqual=${cqual}"
esac

compiler_cmake="${sourcedir}/${pkgtarname}/cmake/compilers/${cid}.cmake"

if (( WANT_TESTS )); then
  tbb_test=ON
else
  tbb_test=OFF
fi

if [ -f "${compiler_cmake}~" ]; then
  cp -p "${compiler_cmake}"{'~',}
else
  cp -p "${compiler_cmake}"{,'~'}
fi

if (( want_ninja )); then
  cmake_options+=( # Build configurations together.
    -G"Ninja Multi-Config"
    "-DCMAKE_CONFIGURATION_TYPES=Release;Debug"
    "-DCMAKE_CROSS_CONFIGS=Release;Debug"
    "-DCMAKE_DEFAULT_CONFIGS=Release;Debug"
  )
  install_configs=(Release Debug)
  [ "${cqual}" = c7 ] &&  \
    cat >> "${compiler_cmake}" <<EOF
# FNAL-EDIT 2021-03-06 CHG - Clang 7 complains about unused -MMD option.
list(APPEND TBB_WARNING_SUPPRESS -Wno-unused-command-line-argument)
EOF
else
  cmake_options+=(-DCMAKE_BUILD_TYPE=Release)
  install_configs=(Release)
fi
cat >> "${compiler_cmake}" <<EOF
# FNAL-EDIT 2021-03-06 CHG - Need to support older CPU.
list(REMOVE_ITEM TBB_COMMON_COMPILE_FLAGS -mrtm)
EOF

cmake_options+=(
  -DBUILD_SHARED_LIBS:BOOL=ON
  ${cxx:+"-DCMAKE_CXX_COMPILER=${cxx}"}
  ${cxxflg[*]:+"-DCMAKE_CXX_FLAGS=${cxxflg[*]}"}
  ${cxxstd:+"-DCMAKE_CXX_STANDARD=${cxxstd}" \
    "-DCMAKE_CXX_STANDARD_REQUIRED:BOOL=ON"}
  -DCMAKE_INSTALL_LIBDIR=lib # Avoid lib64 on some platforms.
  -DCMAKE_INSTALL_PREFIX="${pkgdir}"
#  -DTBB4PY:BOOL=OFF # ttb4py (no).
#  -DTBB_CPF:BOOL=OFF # Preview features (no).
  -DTBB_EXAMPLES:BOOL=OFF # Examples (not currently provided)
  -DTBB_FIND_PACKAGE:BOOL=OFF # Run tests against an external DB (no).
  -DTBB_STRICT:BOOL=ON
  -DTBB_TEST:BOOL=${tbb_test}
)

##################
# Configure parallelism. *C*
ncores=$(ncores)

[[ -n "${CMAKE_BUILD_PARALLEL_LEVEL}" ]] || \
  export CMAKE_BUILD_PARALLEL_LEVEL=${ncores}

[[ -n "${CTEST_PARALLEL_LEVEL}" ]] || \
  export CTEST_PARALLEL_LEVEL=${ncores}
##################
####################################

####################################
# Configure, build, and install.
#
# Use ${blddir}/${pkgtarname} for in-source builds, ${blddir} otherwise.

${trace} cd "${blddir}" || \
  ssi_die "unable to cd to build directory ${blddir}"

##################
# Configure. *C*
${trace} cmake "${cmake_options[@]}" ${DEBUG_CONFIG:+--verbose} \
  "${sourcedir}/${pkgtarname}" || \
  ssi_die "configure failed"
##################

##################
# Build. *C*
${trace} cmake --build "${blddir}" ${DEBUG_BUILD:+--verbose} || \
  ssi_die "build failed"
##################

##################
# Test. *C*
(( WANT_TESTS )) && \
  { ${trace} ctest --output-on-failure || ssi_die "tests failed"; }
##################

##################
# Install. *C*
for cfg in "${install_configs[@]}"; do
  ${trace} cmake --install "${blddir}" --config ${cfg} || ssi_die "install failed"
done
##################

ssi_info "finished building ${package} ${pkgver}"

####################################
# Pre-declaration processing. *C*
#
# Move things, install from contrib, clean build products from in-source
# builds, etc.
####################################

##################
# Declare and set up the installed package. *C*
#
# Add -c for current declaration if required.
${trace} declare_product -x -- || exit
##################

####################################
# Post-declaration processing. *C*
#
# Operations requiring the *final* declaration and setup (as opposed to
# the pre-build fake declaration and setup). This should be rare.
####################################

#################
# Verify installation and make distribution archive. *C*
#
# To customize verification, see is_built(), since it is also used by
# maybe_make_distribution(). See the documentation for those functions
# (above) for customization opportunities.

# Add a message to ssi_build_failed if appropriate.
${trace} is_built && ${trace} ssi_build_completed && \
  ssi_info "${package} is installed at ${pkgdir}" && \
  ${trace} maybe_make_distribution || \
  ssi_build_failed # *C*
########################################################################
# Nothing beyond this point.
########################################################################
